/*
 * test_ring
 * Copyright (C) 1999  Stefan Keller-Tuberg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2 as published
 * by the Free Software Foundation.
 *
 * These conditions include but are not limited to the requirement that the
 * origianal author is acknowledged in any reused, modified or adapted
 * versions of any part of this code and that any system containing reused or
 * modified parts of this code are issued under the GNU General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ******************************************************************************
 * test_ring.c
 *
 * Stand alone utility to test the integrity of a train controller ring.
 *
 * Written to run under the Linux operating system but could be easily hacked
 * to run under one of Linux's lesser counterparts.
 *
 * Compile using gcc -O2 -o test_ring test_ring.c
 *
 * Needs to be run with root priviledges. Most easily achieved by running
 * (if compiled by a non root user)
 *      chown root test_ring
 *      chgrp root test_ring
 * and then
 *      chmod 4755 test_ring
 */

#include <stdio.h>
#include <unistd.h>           /* for calls to ioperm, sleep, usleep */
#include <sys/io.h>           /* for calls to ioperm */
#include <stdlib.h>

/* Uncomment the following if your compiler gives complaints about missing definitions */
/* #include <asm/io.h>           for inb, outb */


#define  BASEPORT     (0x378)

#define  Cmd_Data_WR  (0x01)
#define  Cmd_Data_RD  (0x10)

#define  Stat_Data_WR (0x02)
#define  Stat_Data_RD (0x20)

#define  Latch_WR     (0x04)
#define  Latch_RD     (0x40)

#define  Clock_WR     (0x08)
#define  Clock_RD     (0x08)

#define  Power_On     (0x20)
#define  Led1         (0x40)
#define  Led2         (0x80)

#define  Ok           (0)
#define  Bad          (-1)

static volatile unsigned char output_contents ;

/******************************************************************************
 * output_diagnostic
 */

void
output_diagnostic (int bits)
{
  int led, in ;

  led = 0 ;

  if (bits & 1)
    led |= Led1 ;

  if (bits & 2)
    led |= Led2 ;

  bits &= 0xf ;
  outb (bits | led | Power_On, BASEPORT) ;

  usleep (100000) ;
  in = inb (BASEPORT+1) ;

  printf ("Out(%1x): %s,%s,%s,%s   In(%02x): %s,%s,%s,%s\n",
	  bits & 0xf,
	  (bits&8) ? "CLOCK":"clock",
	  (bits&4) ? "LATCH":"latch",
	  (bits&2) ? "STAT_DAT":"stat_dat",
	  (bits&1) ? "CMD_DAT":"cmd_dat",
	  in,
	  (in&Clock_RD) ? "CLOCK":"clock",
	  (in&Latch_RD) ? "LATCH":"latch",
	  (in&Stat_Data_RD) ? "STAT_DAT":"stat_dat",
	  (in&Cmd_Data_RD) ? "CMD_DAT":"cmd_dat") ;
}

/******************************************************************************
 * fail_abort
 *
 * This routine is called if it has become apparent that the controller ring
 * is opened or that one of the train controller cards is powered off (so that
 * the ring is effectively open)
 */

void
fail_abort()
{
  fprintf (stderr, "It appears that the ring is open or not functioning\n") ;
  exit (1) ;
}

/******************************************************************************
 * wait_for_high, wait_for_low
 *
 * inline subroutines to wait until a bit propagates around the controller
 * ring and arrives back at the PC.
 *
 * These routines "time" the return of the signal around the loop and return
 * a non zero value if the timer expires. A zero returned value indicates
 * success.
 */

#define Max_Count   (5000)    /* maximum allowable loops before bailing */

__inline__ int
wait_for_high (const unsigned char bits)
{
  register unsigned short counter ;

  for (counter=Max_Count ; counter != 0 ; counter--)
  {
    if (inb(BASEPORT+1) & bits)
      return (Ok) ;
  }

  return (Bad) ;
}

__inline__ int
wait_for_low (const unsigned char bits)
{
  register unsigned short counter ;

  for (counter=Max_Count ; counter != 0 ; counter--)
  {
    if ((inb(BASEPORT+1) & bits) == 0)
      return (Ok) ;
  }

  return (Bad) ;
}

/******************************************************************************
 * confirm_ring_operational
 *
 * This routine checks the basic connectivity of the ring.
 */

void
confirm_ring_operational (void)
{
  int error ;
  unsigned char val ;
  int i, c ;

  /*
   * The ring consists of four bits: two data bits (write and read data), a
   * clock bit and a latch control bit. The clock and latch bits are looped
   * around so that they can be monitored to measure the propagation delay
   * around the ring (for optimising the ring data transfer rate). The data
   * bits are also looped but through arrays of shift registers: the data bits
   * which are written to the ring do not appear back at the read inputs until
   * a certain number of clock pulses have been issued.
   *
   * The first test that the ring is operational is the ability to see that
   * the clock and latch bits propagate around the ring within a short period
   * of time (measured in microseconds).
   */

  error = 0 ;
  outb (output_contents | Latch_WR, BASEPORT) ;
  usleep (200000) ;

  if (((val=inb(BASEPORT+1)) & Latch_RD) == 0)
  {
      fprintf (stderr, "Latch does not appear to be looped - 1 not passed (status port reads %#0x)\n", val) ;
      error = 1 ;
  }

  outb (output_contents | Clock_WR, BASEPORT) ;
  usleep (200000) ;

  if (((val=inb(BASEPORT+1)) & Clock_RD) == 0)
  {
      fprintf (stderr, "Clock does not appear to be looped - 1 not passed (status port reads %#0x)\n", val) ;
      error = 1 ;
  }

  outb (output_contents, BASEPORT) ;
  usleep (200000) ;
  val = inb(BASEPORT+1) ;

  if (val & Latch_RD)
  {
      fprintf (stderr, "Latch does not appear to be looped - 0 not passed (status port reads %#0x)\n", val) ;
      error = 1 ;
  }

  if (val & Clock_RD)
  {
      fprintf (stderr, "Clock does not appear to be looped - 0 not passed (status port reads %#0x)\n", val) ;
      error = 1 ;
  }

  if (error == 0)
  {
    printf ("The ring seems to be operational for the clock and latch lines.\n") ;
  }
  else
  {
    fprintf (stderr, "\n") ;
    fprintf (stderr, "There is a problem with the ring....\n") ;
    fprintf (stderr, "1. Is the parallel port adapter connected to the PC?\n") ;
    fprintf (stderr, "2. Is the port it's connected to located at address %#4x?\n", BASEPORT) ;
    fprintf (stderr, "3. Are the ring cables connected correctly (orientation) and in a ring?\n") ;
    fprintf (stderr, "4. Are all of the train controller cards in the ring powered ON?\n") ;
    fprintf (stderr, "5. The PC interface or one of the train controller cards may be broken.\n") ;
    fprintf (stderr, "\n\nPress <ENTER> to commence diagnostic test, <CTRL-C> to quit\n") ;

    while ((c = getchar()) != '\n') ;

    for (i=0 ; i < 10 ; i++)
      for (c=0 ; c < 16 ; c++)
      {
	output_diagnostic (c) ;
	sleep (1) ;
      }

    exit (1) ;
  }
}

/******************************************************************************
 * measure_ring_length
 *
 * This routine measures the length of the command and status rings (in bits)
 * and displays the result.
 */

#define Large_Number_Of_Bits    (1000)

void
measure_ring_length (void)
{
  unsigned short counter, size_of_ring ;

  /*
   * Shift a long string of zeros out to the data command and status ports
   */

  for (counter=Large_Number_Of_Bits ; counter != 0 ; counter--)
  {
    /*
     * Issue a clock pulse: positive then negative transition.
     */

    output_contents |= Clock_WR ;
    output_contents &= ~Cmd_Data_WR & ~Stat_Data_WR ;
    outb (output_contents, BASEPORT) ;

    if (wait_for_high (Clock_RD))
      fail_abort() ;

    output_contents &= ~Clock_WR ;
    outb (output_contents, BASEPORT) ;

    if (wait_for_low (Clock_RD))
      fail_abort() ;
  }

  /*
   * Now start shifting ones out the command data port and count the number of
   * clocks until they start coming back again.
   */

  output_contents |= Cmd_Data_WR ;
  outb (output_contents, BASEPORT) ;
  usleep (20) ; /* microseconds */
  size_of_ring = 0 ;

  for (counter=Large_Number_Of_Bits ; counter != 0 ; counter--, size_of_ring++)
  {
    if (inb(BASEPORT+1) & Cmd_Data_RD)
      break ;

    /*
     * Issue a clock pulse: positive then negative transition.
     */

    output_contents |= Clock_WR ;
    outb (output_contents, BASEPORT) ;

    if (wait_for_high (Clock_RD))
      fail_abort() ;

    output_contents &= ~Clock_WR ;
    outb (output_contents, BASEPORT) ;

    if (wait_for_low (Clock_RD))
      fail_abort() ;
  }

  if (counter == 0)
  {
    fprintf (stderr, "Error: size of status ring seems to be more than %d bits\n", Large_Number_Of_Bits) ;
    exit (1) ;
  }

  printf ("Size of Command Data ring appears to be %d bits\n", size_of_ring) ;

  /*
   * Now start shifting ones out the status data port and count the number of
   * clocks until they start coming back again.
   */

  output_contents &= ~Cmd_Data_WR ;
  output_contents |= Stat_Data_WR ;
  outb (output_contents, BASEPORT) ;
  usleep (20) ; /* microseconds */
  size_of_ring = 0 ;

  for (counter=Large_Number_Of_Bits ; counter != 0 ; counter--, size_of_ring++)
  {
    if (inb(BASEPORT+1) & Stat_Data_RD)
      break ;

    /*
     * Issue a clock pulse: positive then negative transition.
     */

    output_contents |= Clock_WR ;
    outb (output_contents, BASEPORT) ;

    if (wait_for_high (Clock_RD))
      fail_abort() ;

    output_contents &= ~Clock_WR ;
    outb (output_contents, BASEPORT) ;

    if (wait_for_low (Clock_RD))
      fail_abort() ;
  }

  if (counter == 0)
  {
    fprintf (stderr, "Error: size of status ring seems to be more than %d bits\n", Large_Number_Of_Bits) ;
    exit (1) ;
  }

  printf ("Size of Status Data ring appears to be %d bits\n", size_of_ring) ;
}

/******************************************************************************
 * Here is where the programme starts. Remember, the executable needs to be
 * SUID root or it needs to be run by the root user.
 *
 * Firstly, claim the printer port, then do some checks to see if the parallel
 * port train controller interface adapter is connected, then run the test to
 * see how long the ring is.
 */

int main()
{
  puts ("test_ring - Copyright (C) 1999 Stefan Keller-Tuberg\n") ;
  puts ("This program is free software; you can redistribute it and/or modify") ;
  puts ("it under the terms of the GNU General Public License Version 2 as published") ;
  puts ("by the Free Software Foundation.\n") ;

  if (ioperm(BASEPORT, 3, 1))
  {
    perror ("ioperm call failed to open parallel port") ;
    puts ("Have you ensured that the programme is run as or SUID root?") ;
    exit (1) ;
  }

  /*
   * Ensure that the clock and latch bits are OFF
   * Also ensure that the power control bit is ON (to enable the ring)
   */

  output_contents = Power_On ;
  outb (output_contents, BASEPORT) ;
  sleep (1) ; /* second */

  confirm_ring_operational() ;
  measure_ring_length() ;

  if (ioperm(BASEPORT, 3, 0))
  {
    perror ("ioperm call failed to close parallel port") ;
    exit (1) ;
  }
  exit (0) ;
}
